package uk.ac.starlink.votable.tools;

import java.io.BufferedOutputStream;
import java.io.BufferedWriter;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import org.w3c.dom.Attr;
import org.w3c.dom.Comment;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.Text;
import uk.ac.starlink.table.StarTable;
import uk.ac.starlink.votable.DataFormat;
import uk.ac.starlink.votable.TableElement;
import uk.ac.starlink.votable.VOSerializer;
import uk.ac.starlink.votable.VOStarTable;

/**
 * Serializes a DOM to XML.  It basically does what you'd expect, except
 * for DATA elements, which are written in accordance with this writer's
 * {@link DataFormat} variable.
 * Note you must remember to close it after use.
 *
 * @author   Mark Taylor (Starlink)
 * @since    22 Sep 2004
 */
class DOMWriter {

    private final BufferedWriter out_;
    private final DataFormat format_;
    private final File base_;
    private final static int INDENT = 2;
    private int iTable_ = 0;

    /** Character encoding for XML data. */
    public static Charset ENCODING = Charset.forName( "UTF-8" );

    /**
     * Constructs a new DOMWriter.
     *
     * @param   out  destination stream for XML
     * @param   format  controls which format variant any TABLE DATA elements
     *          will be written in
     * @param   base  base name for external href files created - 
     *          if null all streams will be inline
     */
    public DOMWriter( OutputStream out, DataFormat format, File base ) {
        out_ = new BufferedWriter( new OutputStreamWriter( out ) );
        format_ = format;
        base_ = base;
    }

    /**
     * Signals that no more nodes will be written.
     * This method must be called to flush the output.
     */
    public void close() throws IOException {
        out_.write( "\n" );
        out_.flush();
    }

    /**
     * Outputs a node.
     * 
     * @param  node to write
     */
    public void writeNode( Node node ) throws IOException {
        if ( node instanceof Document ) {
            writeDocument( (Document) node );
        }
        else if ( node instanceof Element ) {
            writeElement( (Element) node );
        }
        else if ( node instanceof Text ) {
            writeText( (Text) node );
        }
        else if ( node instanceof Comment ) {
            writeComment( (Comment) node );
        }
    }

    public void writeDocument( Document doc ) throws IOException {
        out_.write( "<?xml" );
        out_.write( VOSerializer.formatAttribute( "version", "1.0" ) );
        out_.write( VOSerializer.formatAttribute( "encoding",
                                                  ENCODING.name() ) );
        out_.write( "?>" );
        writeElement( doc.getDocumentElement() );
    }

    public void writeText( Text text ) throws IOException {
        out_.write( text.getData() );
    }

    public void writeComment( Comment comment ) throws IOException {
        out_.write( "<!-- " );
        out_.write( comment.getData() );
        out_.write( "-->" );
    }

    public void writeElement( Element el ) throws IOException {

        /* Treat a DATA child of a TABLE specially. */
        if ( el.getTagName().equals( "DATA" ) &&
             el.getParentNode() instanceof TableElement ) {
            writeData( (TableElement) el.getParentNode() );
            return;
        }

        /* Obtain the interesting children. */
        List children = new ArrayList();
        int nEl = 0;
        int nOther = 0;
        for ( Node ch = el.getFirstChild(); ch != null;
              ch = ch.getNextSibling() ) {
            if ( ch instanceof Element ) {
                nEl++;
                children.add( ch );
            }
            else if ( ch instanceof Text ) {
                nOther++;
                children.add( ch );
            }
        }

        /* Get indentation prefix. */
        String prefix;
        if ( INDENT > 0 ) {
            int level = 0;
            for ( Node ancestor = el.getParentNode(); 
                  ancestor instanceof Element; 
                  ancestor = ancestor.getParentNode() ) {
                level++;
            }
            char[] spaces = new char[ level * INDENT ];
            Arrays.fill( spaces, ' ' );
            prefix = new String( spaces );
        }
        else {
            prefix = "";
        }

        /* Open start tag. */
        out_.write( '\n' );
        out_.write( prefix );
        out_.write( '<' );
        out_.write( el.getTagName() );

        /* Attributes. */
        NamedNodeMap atts = el.getAttributes();
        if ( atts != null ) {
            for ( int i = 0; i < atts.getLength(); i++ ) {
                Attr att = (Attr) atts.item( i );
                out_.write( VOSerializer.formatAttribute( att.getName(),
                                                          att.getValue() ) );
            }
        }

        /* Close start tag. */
        if ( children.isEmpty() ) {
            out_.write( "/>" );
        }
        else {
            out_.write( ">" );

            /* Recursively process child elements. */
            for ( Iterator it = children.iterator(); it.hasNext(); ) {
                writeNode( (Node) it.next() );
            }

            /* End tag. */
            if ( nEl > 0 ) {
                out_.write( '\n' );
                out_.write( prefix );
            }
            out_.write( "</" );
            out_.write( el.getTagName() );
            out_.write( ">" );
        }
    }

    private void writeData( TableElement tableEl ) throws IOException {
        iTable_++;
        if ( format_ != null ) {
            out_.write( '\n' );
            StarTable st = new VOStarTable( tableEl );
            VOSerializer vser = VOSerializer.makeSerializer( format_, st );
            if ( base_ == null || format_ == DataFormat.TABLEDATA ) {
                vser.writeInlineDataElement( out_ );
            }
            else {
                String suffix;
                if ( format_ == DataFormat.FITS ) {
                    suffix = ".fits";
                }
                else if ( format_ == DataFormat.BINARY ) {
                    suffix = ".bin";
                }
                else {
                    suffix = null;
                    assert false;
                }
                File dest = new File( base_.getPath() + "-" + iTable_ 
                                    + suffix );
                DataOutputStream dataOut = 
                    new DataOutputStream(
                        new BufferedOutputStream(
                            new FileOutputStream( dest ) ) );
                vser.writeHrefDataElement( out_, dest.getPath(), dataOut );
                dataOut.flush();
                dataOut.close();
            }
        }
    }

}
